#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "lex.h"

/* lexical analysis routine
   returns LETTER for a lower case letter, .i = 0 through 25
   return NUMBER for a for numbers, converted double in .d
   all other characters are returned immediately */
int calc_lex(calc_ctx_t *ctx, calc_STYPE *yylval)
{
	int c, n;
	char buf[128];
	while((c=getchar()) == ' ') ctx->col++; /* skip blanks */

	ctx->col++;

	if (c == '\n') {
		ctx->col = 0;
		ctx->line++;
	}

	yylval->line = ctx->line;
	yylval->first_col = yylval->last_col = ctx->col;

	/* c is now nonblank */
	if (islower(c)) {
		yylval->un.i = c - 'a';
		return LETTER;
	}

	/* collect numbers in tmp[] and use strtod() to convert them; this is
	   an example on tracking first-last location of tokens. To see this
	   in action, input: 1+2.3.4+5: it will say line:6..7 caused the error,
	   which is ".4" because 2.3 is a valid numeric token and .4 is a valid
	   numeric token too. */
	if (isdigit(c) || (c == '.')) {
		int seen_e = 0, seen_esign = 0, seen_dot = 0;
		buf[0] = c;
		for(n = 1; n < sizeof(buf)-1; n++) {
			c = getchar();
			ctx->col++;
			if (c == '.') {
				if (seen_dot)
					break;
				seen_dot = 1;
			}
			if ((c == 'e') || (c == 'E')) {
				if (seen_e)
					break;
				seen_e = 1;
			}
			if ((c == '+') || (c == '-')) {
				if ((!seen_e) || (seen_esign))
					break;
				seen_esign = 1;
			}
			if (isdigit(c) || (c == '+') || (c == '-') || (c == 'e') || (c == 'E') || (c == '.'))
				buf[n] = c;
			else
				break;
		}
		ungetc(c, stdin);
		ctx->col--;
		yylval->last_col = ctx->col;
		buf[n] = '\0';
		yylval->un.d = strtod(buf, NULL);
		return NUMBER;
	}
	return c;
}


