%prefix calc_%

%{
# include <stdio.h>
# include <ctype.h>

typedef struct {
	int regs[26];
	int base;
} calc_ctx_t;

%}

%start list

%token DIGIT LETTER

%left '|'
%left '&'
%left '+' '-'
%left '*' '/' '%'
%left UMINUS   /* supplies precedence for unary minus */

%% /* beginning of rules section */

list  :  /* empty */
      |  list stat '\n'
      |  list error '\n'
            {  calc_errok ; }
      ;

stat  :  expr
            {  printf("%d\n",$1);}
      |  LETTER '=' expr
            {  ctx->regs[$1] = $3; }
      ;

expr  :  '(' expr ')'
            {  $$ = $2; }
      |  expr '+' expr
            {  $$ = $1 + $3; }
      |  expr '-' expr
            {  $$ = $1 - $3; }
      |  expr '*' expr
            {  $$ = $1 * $3; }
      |  expr '/' expr
            {  $$ = $1 / $3; }
      |  expr '%' expr
            {  $$ = $1 % $3; }
      |  expr '&' expr
            {  $$ = $1 & $3; }
      |  expr '|' expr
            {  $$ = $1 | $3; }
      |  '-' expr %prec UMINUS
            {  $$ = - $2; }
      |  LETTER
            {  $$ = ctx->regs[$1]; }
      |  number
      ;

number:  DIGIT
         {  $$ = $1; ctx->base = ($1==0) ? 8 : 10; }
      |  number DIGIT
         {  $$ = ctx->base * $1 + $2; }
      ;

%% /* start of programs */

static int lex(calc_ctx_t *ctx, calc_STYPE *yylval)
{
	/* lexical analysis routine */
	/* returns LETTER for a lower case letter, *yylval = 0 through 25 */
	/* return DIGIT for a digit, *yylval = 0 through 9 */
	/* all other characters are returned immediately */

	int c;
	while( (c=getchar()) == ' ' )   { /* skip blanks */ }

	/* c is now nonblank */
	if (islower(c)) {
		*yylval = c - 'a';
		return LETTER;
	}
	if (isdigit(c)) {
		*yylval = c - '0';
		return DIGIT;
	}
	return(c);
}

int main(void)
{
	calc_ctx_t ctx;
	calc_yyctx_t yyctx;

	if (calc_parse_init(&yyctx) != 0)
		return -1;

	for(;;) {
		int tok, res;
		calc_STYPE lval;

		if (feof(stdin)) {
			printf("<feof>\n");
			break;
		}
		tok = lex(&ctx, &lval);
		res = calc_parse(&yyctx, &ctx, tok, &lval);
	}
	return 0;
}

void calc_error(calc_ctx_t *ctx, calc_STYPE tok, const char *s)
{
	fprintf(stderr, "%s\n", s);
}

