#include "defs.h"

/* TABLE_SIZE is the number of entries in the symbol table. */
/* TABLE_SIZE must be a power of two.			    */

#define	TABLE_SIZE 1024

hashtbl_t symtab;

static int hash(const char *name)
{
	const char *s;
	int c, k;

	assert(name && *name);
	s = name;
	k = *s;
	while ((c = *++s) != 0)
		k = (31 * k + c) & (TABLE_SIZE - 1);

	return (k);
}

bucket *make_bucket(const char *name)
{
	bucket *bp;

	assert(name != 0);

	bp = TMALLOC(bucket, 1);
	NO_SPACE(bp);

	bp->link = 0;
	bp->next = 0;

	bp->name = TMALLOC(char, strlen(name) + 1);
	NO_SPACE(bp->name);

	bp->tag = 0;
	bp->value = UNDEFINED;
	bp->index = 0;
	bp->prec = 0;
	bp->class = UNKNOWN;
	bp->assoc = TOKEN;
#if defined(YYBTYACC)
	bp->args = -1;
	bp->argnames = 0;
	bp->argtags = 0;
	bp->destructor = 0;
#endif
	strcpy(bp->name, name);

	return (bp);
}

bucket *lookup(hashtbl_t *ht, const char *name, int alloc)
{
	bucket *bp, **bpp;

	bpp = ht->symbol_table + hash(name);
	bp = *bpp;

	while (bp) {
		if (strcmp(name, bp->name) == 0)
			return (bp);
		bpp = &bp->link;
		bp = *bpp;
	}

	if (!alloc)
		return NULL;

	*bpp = bp = make_bucket(name);
	ht->last_symbol->next = bp;
	ht->last_symbol = bp;

	return (bp);
}

void create_symbol_table(hashtbl_t *ht)
{
	int i;

	memset(ht, 0, sizeof(hashtbl_t));
	ht->symbol_table = TMALLOC(bucket *, TABLE_SIZE);
	NO_SPACE(ht->symbol_table);

	for (i = 0; i < TABLE_SIZE; i++)
		ht->symbol_table[i] = 0;

}

void symtab_insert_first(hashtbl_t *ht, char *name)
{
	bucket *bp;

	bp = make_bucket(name);
	bp->index = 1;
	bp->class = TERM;

	ht->first_symbol = bp;
	ht->last_symbol = bp;
	ht->symbol_table[hash(name)] = bp;
}


void free_symbol_table(hashtbl_t *ht)
{
	FREE(ht->symbol_table);
	ht->symbol_table = 0;
}

void free_symbols(hashtbl_t *ht)
{
	bucket *p, *q;

	for (p = ht->first_symbol; p; p = q) {
		q = p->next;
		FREE(p);
	}
}
